const version = "0.0.1-beta.1";

function concat(datas) {
    const length = datas.reduce((a, d) => (a + d.length), 0);
    const result = new Uint8Array(length);
    let offset = 0;
    for (const data of datas) {
        result.set(data, offset);
        offset += data.length;
    }
    return result;
}
function stall(duration) {
    return new Promise((resolve) => {
        setTimeout(resolve, duration);
    });
}
function fromLeBytes(bytes) {
    let result = 0;
    for (let i = 0; i < bytes.length; i++) {
        result |= (bytes[i] << (i << 3));
    }
    return result;
}
function toLeBytes(value, width) {
    const result = new Uint8Array(width);
    for (let i = 0; i < width; i++) {
        result[i] = value & 0xff;
        value >>= 8;
    }
    return result;
}
const HEX = "0123456789abcdef";
function hexlify(value, width) {
    let result = "";
    if (typeof (value) === "number") {
        result = value.toString(16);
    }
    else {
        for (let i = 0; i < value.length; i++) {
            const d = value[i];
            result += HEX[d >> 4] + HEX[d & 0x0f];
        }
    }
    if (typeof (width) === "number") {
        while (result.length < 2 * width) {
            result = "0" + result;
        }
    }
    else if (width) {
        if (result.length % 2) {
            result = "0" + result;
        }
    }
    return result;
}

function pad(c, length) {
    while (c.length < length) {
        c += c;
    }
    return c.substring(0, length);
}
class Logger {
    printer;
    #indent;
    constructor(printer) {
        this.printer = printer || console.log;
        this.#indent = 0;
    }
    get indent() { return this.#indent; }
    data(data, title) {
        if (title == null) {
            title = "[DATA]";
        }
        if (title) {
            title += " ";
        }
        this._print(`${title}${data.length} bytes`);
        for (let i = 0; i < data.length; i += 16) {
            const chunk = data.slice(i, i + 16);
            let line = "  " + hexlify(chunk.slice(0, 8)) + " " + hexlify(chunk.slice(8));
            line += pad(" ", 37 - line.length) + "|  ";
            for (let j = 0; j < chunk.length; j++) {
                const c = chunk[j];
                if (c >= 32 && c <= 127) {
                    line += String.fromCharCode(c);
                }
                else {
                    line += ".";
                }
            }
            this._print(line);
        }
    }
    info(message) {
        this._print(message);
    }
    object(obj, title) {
        if (title == null) {
            title = "[DATA]";
        }
        if (title) {
            title += " ";
        }
        this._print(`${title}${JSON.stringify(obj)}`);
    }
    _print(message) {
        if (message.startsWith("<<|")) {
            message = message.substring(3).trim();
            this.#indent--;
        }
        let indent = message.startsWith("|>>");
        if (indent) {
            message = message.substring(3).trim();
        }
        this.printer(pad(" ", this.#indent * 2) + message);
        if (indent) {
            this.#indent++;
        }
    }
    static voidLogger() {
        return new Logger((l) => { });
    }
    static logger() {
        return new Logger();
    }
}

const ESP_READ_REG = 0x0a;

function slipEncode(data) {
    let stuffBytes = 0;
    for (let i = 0; i < data.length; i++) {
        let c = data[i];
        if (c === 0xdb || c === 0xc0) {
            stuffBytes++;
        }
    }
    const slipData = new Uint8Array(2 + stuffBytes + data.length);
    let offset = 0;
    slipData[offset++] = 0xc0;
    for (let i = 0; i < data.length; i++) {
        let c = data[i];
        if (c === 0xdb) {
            slipData[offset++] = 0xdb;
            slipData[offset++] = 0xdd;
        }
        else if (c === 0xc0) {
            slipData[offset++] = 0xdb;
            slipData[offset++] = 0xdc;
        }
        else {
            slipData[offset++] = c;
        }
    }
    slipData[offset++] = 0xc0;
    return slipData;
}
function slipDecode(data) {
    const markers = [];
    for (let i = 0; i < data.length; i++) {
        if (data[i] === 0xc0) {
            markers.push(i);
        }
        if (markers.length < 2) {
            continue;
        }
        const result = [];
        for (let i = markers[0] + 1; i < markers[1]; i++) {
            if (data[i] === 0xdb) {
                if (data[i + 1] === 0xdc) {
                    result.push(0xc0);
                    i++;
                    continue;
                }
                else if (data[i + 1] === 0xdd) {
                    result.push(0xdb);
                    i++;
                    continue;
                }
            }
            result.push(data[i]);
        }
        return {
            data: new Uint8Array(result),
            start: markers[0],
            consumed: markers[1] + 1
        };
    }
    return null;
}

// @TOOD: Move to constants file?
const ROM_INVALID_RECV_MSG = 0x05;
function getValue(value, fallback) {
    if (value == null) {
        return fallback;
    }
    if (value < -1) {
        return -1;
    }
    if (!Number.isInteger(value)) {
        throw new TypeError(`Invalid integer range: ${value}`);
    }
    return value;
}
function sum(values) {
    return values.reduce((accum, value) => (accum + value.length), 0);
}
const Sequences = {
    Reset: "R 100 D 50 N",
    ResetUsb: "N 100 D 100 R 100 R 100 N",
    ResetHard: "R 100 N",
};
let SerialPort$1 = class SerialPort {
    #listeners;
    #readBuffer;
    #writeBuffer;
    #maxReadBuffer;
    #maxWriteBuffer;
    logger;
    constructor(options) {
        if (options == null) {
            options = {};
        }
        this.#listeners = [];
        this.#readBuffer = [];
        this.#writeBuffer = [];
        this.#maxReadBuffer = getValue(options.maxReadBuffer, 1 << 20);
        this.#maxWriteBuffer = getValue(options.maxReadBuffer, 1 << 20);
        this.logger = options.logger || Logger.logger();
    }
    get maxReadBuffer() { return this.#maxReadBuffer; }
    get maxWriteBuffer() { return this.#maxWriteBuffer; }
    get available() { return sum(this.#readBuffer); }
    get backlog() { return sum(this.#writeBuffer); }
    /**
     *  Reads any data on the stream into the read buffer and returns
     *  the combined data.
     *
     *  Use [[_unread]] to place any data back on the read buffer
     *  to be processed in the future.
     */
    async read() {
        this.#readBuffer.push(await this._read());
        const result = concat(this.#readBuffer);
        this.#readBuffer = [];
        return result;
    }
    /**
     *  Read a SLIP packet, optionally matching the %%op%%. Returns
     *  ``null`` if no complete matching packet is found.
     *
     *  Any stray packets or bytes at the front of the read buffer
     *  are discarded.
     */
    async readSlipPacket(op) {
        const data = await this.read();
        const slip = slipDecode(data);
        // No packet found; maybe we need more bytes
        if (slip == null) {
            this._unread(data);
            return null;
        }
        // Place unconsumed bytes back onto the read buffer
        this._unread(data.slice(slip.consumed));
        const resp = slip.data[0];
        const opRet = slip.data[1];
        const value = fromLeBytes(slip.data.slice(4, 8));
        const result = slip.data.slice(8);
        if (resp === 1) {
            if (op == null || opRet === op) {
                return { value, data: result };
            }
            if (result[0] !== 0 && result[1] === ROM_INVALID_RECV_MSG) {
                // @TODO: flush input?
                throw new Error("unsupported command");
            }
        }
        throw new Error(`Bad response: ${slip}`);
    }
    _unread(data) {
        this.#readBuffer.unshift(data);
    }
    async write(data) {
        return this._write(data);
    }
    async writeSlipPacket(data) {
        return this.write(slipEncode(data));
    }
    /**
     *  Reset the connected device, optionally providing a reset
     *  %%sequence%% for fine control over the DTR/RTS signals.
     */
    async reset(sequence) {
        if (sequence == null) {
            sequence = Sequences.ResetUsb;
        }
        this.logger.info(`Reset: ${sequence}`);
        for (const cmd of sequence.split(/ /g)) {
            switch (cmd) {
                case "N":
                    await this.signal({});
                    break;
                case "D":
                    await this.signal({ dtr: true });
                    break;
                case "R":
                    await this.signal({ rts: true });
                    break;
                case "DR":
                case "RD":
                    await this.signal({ dtr: true, rts: true });
                    break;
                default:
                    await stall(parseInt(cmd));
            }
        }
    }
    /**
     *  Send a command to the connected device and parse the response.
     */
    async command(op, data, checksum) {
        if (Array.isArray(data)) {
            data = new Uint8Array(data);
        }
        this.logger.info(`|>> Command: op=0x${hexlify(op, 1)} data=0x${data ? hexlify(data) : ""}`);
        const packet = new Uint8Array(8 + (data ? data.length : 0));
        packet[0] = 0x00;
        packet[1] = op;
        if (data) {
            packet.set(toLeBytes(data.length, 2), 2);
            packet.set(data, 8);
        }
        if (checksum) {
            packet.set(toLeBytes(checksum, 4), 4);
        }
        await this.writeSlipPacket(packet);
        for (let i = 0; i < 100; i++) {
            const result = await this.readSlipPacket(op);
            if (result) {
                this.logger.info(`<<| (Command) value=0x${hexlify(result.value, 4)} data=${hexlify(result.data)}`);
                return result;
            }
            await stall(10);
        }
        this.logger.info(`<<| (Command) null`);
        return null;
    }
    async readRegister(address) {
        this.logger.info(`|>> Register: address=${hexlify(address, 4)}`);
        const packet = toLeBytes(address, 4);
        const { value } = await this.command(ESP_READ_REG, packet);
        this.logger.info(`<<| (Register) value=${hexlify(value, 4)}`);
        return value;
    }
    async sync() {
        this.logger.info(`|>> Sync:`);
        const packet = new Uint8Array(36);
        packet[0] = 0x07;
        packet[1] = 0x07;
        packet[2] = 0x12;
        packet[3] = 0x20;
        packet.fill(0x55, 4);
        const result = await this.command(0x08, packet);
        this.logger.info(`<<| (Sync) value=${hexlify(result, 4)}`);
        return result;
    }
    addListener(event, listener) {
        this.#listeners.push({ event, listener });
        return this;
    }
    listeners(event) {
        return this.#listeners.reduce((accum, l) => {
            if (l.event === event) {
                accum.push(l.listener);
            }
            return accum;
        }, []);
    }
    removeAllListeners(event) {
        if (event == null) {
            this.#listeners = [];
            return this;
        }
        this.#listeners = this.#listeners.filter((l) => {
            return (l.event !== event);
        });
        return this;
    }
    removeListener(event, listener) {
        if (listener == null) {
            return this.removeAllListeners(event);
        }
        else {
            let found = false;
            this.#listeners = this.#listeners.filter((l) => {
                if (found || l.event !== event) {
                    return true;
                }
                if (listener === l.listener) {
                    found = true;
                    return false;
                }
                return true;
            });
        }
        return this;
    }
    once(event, listener) {
        this.#listeners.push({ event, listener, once: true });
        return this;
    }
    emit(event, ...args) {
        // Recompute the remaining listeners, tracking valid targets
        const listeners = [];
        this.#listeners = this.#listeners.filter((listener) => {
            if (listener.event !== event) {
                return true;
            }
            listeners.push(listener.listener);
            return !listener.once;
        });
        // Synchronously notify each
        for (const listener of listeners) {
            try {
                listener(...args);
            }
            catch (error) { }
        }
        return listeners.length > 0;
    }
    on(event, listener) {
        return this.addListener(event, listener);
    }
    off(event, listener) {
        return this.removeListener(event, listener);
    }
    listenerCount(event) {
        return this.#listeners.reduce((accum, l) => (accum += (l.event === event ? 1 : 0)), 0);
    }
};

class SerialPort extends SerialPort$1 {
    port;
    constructor(port) {
        super();
        this.port = port;
    }
    async signal(signal) {
    }
    async getSignal() {
        return { dtr: false, rts: false };
    }
    async _write(data) {
        return false;
    }
    async _read() {
        return new Uint8Array([]);
    }
}

export { Logger, Sequences, SerialPort$1 as SerialPort, SerialPort as SerialPortBrowser, slipDecode, slipEncode, version };
//# sourceMappingURL=firefly.js.map
