"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Device = exports.Sequences = void 0;
const constants_js_1 = require("./constants.js");
const logger_js_1 = require("./logger.js");
const slip_js_1 = require("./slip.js");
const utils_js_1 = require("./utils.js");
// @TOOD: Move to constants file?
const ROM_INVALID_RECV_MSG = 0x05;
exports.Sequences = {
    Reset: "R 100 D 50 N",
    ResetUsb: "N 100 D 100 R 100 R 100 N",
    ResetHard: "R 100 N",
};
;
class Device {
    serial;
    #readBuffer;
    #writeBuffer;
    #maxReadBuffer;
    #maxWriteBuffer;
    logger;
    constructor(serial, options) {
        if (options == null) {
            options = {};
        }
        this.serial = serial;
        this.logger = options.logger || logger_js_1.Logger.logger();
        this.#readBuffer = [];
        this.#writeBuffer = [];
        this.#maxReadBuffer = getValue("invalid options.maxReadBuffer", options.maxReadBuffer, 1 << 20);
        this.#maxWriteBuffer = getValue("invalid options.maxWriteBuffer", options.maxReadBuffer, 1 << 20);
    }
    get maxReadBuffer() { return this.#maxReadBuffer; }
    get maxWriteBuffer() { return this.#maxWriteBuffer; }
    get available() { return sum(this.#readBuffer); }
    get backlog() { return sum(this.#writeBuffer); }
    /**
     *  Reads any data on the stream into the read buffer and returns
     *  the combined data.
     *
     *  Use [[_unread]] to place any data back on the read buffer
     *  to be processed in the future.
     */
    async read() {
        const input = await this.serial.read();
        if (input.length) {
            this.logger.data(input, "<<<");
            this.#readBuffer.push(input);
        }
        const result = (0, utils_js_1.concat)(this.#readBuffer);
        this.#readBuffer = [];
        return result;
    }
    /**
     *  Read a SLIP packet, optionally matching the %%op%%. Returns
     *  ``null`` if no complete matching packet is found.
     *
     *  Any stray packets or bytes at the front of the read buffer
     *  are discarded.
     */
    async readSlipPacket(op) {
        const data = await this.read();
        const slip = (0, slip_js_1.slipDecode)(data);
        // No packet found; maybe we need more bytes
        if (slip == null) {
            this._unread(data);
            return null;
        }
        // Place unconsumed bytes back onto the read buffer
        this._unread(data.slice(slip.consumed));
        const resp = slip.data[0];
        const opRet = slip.data[1];
        const value = (0, utils_js_1.fromLeBytes)(slip.data.slice(4, 8));
        const result = slip.data.slice(8);
        if (resp === 1) {
            if (op == null || opRet === op) {
                return { value, data: result };
            }
            if (result[0] !== 0 && result[1] === ROM_INVALID_RECV_MSG) {
                // @TODO: flush input?
                throw new Error("unsupported command");
            }
        }
        throw new Error(`Bad response: ${slip}`);
    }
    _unread(data) {
        this.#readBuffer.unshift(data);
    }
    async write(data) {
        this.logger.data(data, ">>>");
        return this.serial.write(data);
    }
    async writeSlipPacket(data) {
        return this.write((0, slip_js_1.slipEncode)(data));
    }
    /**
     *  Reset the connected device, optionally providing a reset
     *  %%sequence%% for fine control over the DTR/RTS signals.
     */
    async reset(sequence) {
        if (sequence == null) {
            sequence = exports.Sequences.ResetUsb;
        }
        this.logger.info(`Reset: ${sequence}`);
        for (const cmd of sequence.split(/ /g)) {
            switch (cmd) {
                case "N":
                    await this.serial.signal({});
                    break;
                case "D":
                    await this.serial.signal({ dtr: true });
                    break;
                case "R":
                    await this.serial.signal({ rts: true });
                    break;
                case "DR":
                case "RD":
                    await this.serial.signal({ dtr: true, rts: true });
                    break;
                default:
                    await (0, utils_js_1.stall)(parseInt(cmd));
            }
        }
    }
    /**
     *  Send a command to the connected device and parse the response.
     */
    async command(op, data, checksum) {
        if (Array.isArray(data)) {
            data = new Uint8Array(data);
        }
        this.logger.info(`|>> Command: op=0x${(0, utils_js_1.hexlify)(op, 1)} data=0x${data ? (0, utils_js_1.hexlify)(data) : ""}`);
        const packet = new Uint8Array(8 + (data ? data.length : 0));
        packet[0] = 0x00;
        packet[1] = op;
        if (data) {
            packet.set((0, utils_js_1.toLeBytes)(data.length, 2), 2);
            packet.set(data, 8);
        }
        if (checksum) {
            packet.set((0, utils_js_1.toLeBytes)(checksum, 4), 4);
        }
        await this.writeSlipPacket(packet);
        for (let i = 0; i < 100; i++) {
            const result = await this.readSlipPacket(op);
            if (result) {
                this.logger.info(`<<| (Command) value=0x${(0, utils_js_1.hexlify)(result.value, 4)} data=${(0, utils_js_1.hexlify)(result.data)}`);
                return result;
            }
            await (0, utils_js_1.stall)(10);
        }
        this.logger.info(`<<| (Command) null`);
        return null;
    }
    async readRegister(address) {
        this.logger.info(`|>> Register: address=${(0, utils_js_1.hexlify)(address, 4)}`);
        const packet = (0, utils_js_1.toLeBytes)(address, 4);
        const { value } = await this.command(constants_js_1.ESP_READ_REG, packet);
        this.logger.info(`<<| (Register) value=${(0, utils_js_1.hexlify)(value, 4)}`);
        return value;
    }
    async sync() {
        this.logger.info(`|>> Sync:`);
        const packet = new Uint8Array(36);
        packet[0] = 0x07;
        packet[1] = 0x07;
        packet[2] = 0x12;
        packet[3] = 0x20;
        packet.fill(0x55, 4);
        const result = await this.command(0x08, packet);
        this.logger.info(`<<| (Sync) value=${(0, utils_js_1.hexlify)(result, 4)}`);
        return result;
    }
}
exports.Device = Device;
function getValue(message, value, fallback) {
    if (value == null) {
        return fallback;
    }
    if (!Number.isInteger(value)) {
        throw new TypeError(`${message}: ${value}`);
    }
    if (value < -1) {
        return -1;
    }
    return value;
}
function sum(values) {
    return values.reduce((accum, value) => (accum + value.length), 0);
}
//# sourceMappingURL=device.js.map