/**
 *  A simple communication library for the ESP Devices over the
 *  Serial Protocol for the UART bootloader.
 *
 *  See:
 *    ESPTool:
 *    ESPTool-js:
 *    Protocol: https://docs.espressif.com/projects/esptool/en/latest/esp32/advanced-topics/serial-protocol.html
 */
import { Logger } from "./logger.js";
import type { SerialPort } from "./serial.js";
export declare const Sequences: Record<string, string>;
export interface DeviceInfo {
    packageVersion: number;
    revision: number;
    description: string;
}
export interface DeviceOptions {
    maxReadBuffer?: number;
    maxWriteBuffer?: number;
    logger?: Logger;
}
export type DeviceResponse = {
    value: number;
    data: Uint8Array;
};
/**
 *  The **BaseDevice** class is a minimal implementation of
 *  of the Serial Protocol necessary to read/write and detect
 *  the device magic number.
 */
export declare class BaseDevice {
    #private;
    readonly serial: SerialPort;
    readonly logger: Logger;
    constructor(serial: SerialPort, options?: DeviceOptions);
    get maxReadBuffer(): number;
    get maxWriteBuffer(): number;
    get available(): number;
    get backlog(): number;
    connect(): Promise<number>;
    /**
     *  Reads any data on the stream into the read buffer and returns
     *  the combined data.
     *
     *  Use [[_unread]] to place any data back on the read buffer
     *  to be processed in the future.
     */
    read(): Promise<Uint8Array>;
    /**
     *  Read a SLIP packet, optionally matching the %%op%%. Returns
     *  ``null`` if no complete matching packet is found.
     *
     *  Any stray packets or bytes at the front of the read buffer
     *  are discarded.
     */
    readSlipPacket(op?: number): Promise<null | DeviceResponse>;
    _unread(data: Uint8Array): void;
    write(data: Uint8Array): Promise<boolean>;
    writeSlipPacket(data: Uint8Array): Promise<boolean>;
    /**
     *  Reset the connected device, optionally providing a reset
     *  %%sequence%% for fine control over the DTR/RTS signals.
     */
    reset(sequence?: string): Promise<void>;
    sync(): Promise<any>;
    /**
     *  Send a command to the connected device and parse the response.
     */
    command(op: number, data?: Array<number> | Uint8Array, checksum?: number): Promise<null | DeviceResponse>;
    readRegister(address: number): Promise<number>;
}
export declare abstract class Device extends BaseDevice {
    connect(): Promise<number>;
    spiFlashCommand(command: number, data: Uint8Array, responseBits: number): Promise<number>;
    readSpiRegister(offset: number): Promise<number>;
    writeSpiRegister(offset: number, value: number, mask?: number, delayUs?: number, delayAfterUs?: number): Promise<void>;
    /**
     *  Returns true if this **Device** supports the %%magic%%
     *  number. This is used in connect to validate the expected
     *  devices matches the connected device.
     */
    abstract checkMagic(magic: number): boolean;
    abstract getDeviceInfo(): Promise<DeviceInfo>;
    abstract registerAddress(register: number): number;
    abstract readonly chipName: string;
    abstract readonly SPI_USR_OFFS: number;
    abstract readonly SPI_USR2_OFFS: number;
    abstract readonly SPI_W0_OFFS: number;
    abstract _setDataLengths(mosiLength: number, misoLength: number): Promise<void>;
    abstract readonly UART_DATE_REG_ADDR: number;
}
//# sourceMappingURL=device.d.ts.map