/**
 *  Serial Line Internet Protocol (SLIP) coder library.
 *
 *  See: https://en.wikipedia.org/wiki/Serial_Line_Internet_Protocol
 */
/**
 *  Encode %%data%% using SLIP encoding.
 */
export function slipEncode(data) {
    let stuffBytes = 0;
    for (let i = 0; i < data.length; i++) {
        let c = data[i];
        if (c === 0xdb || c === 0xc0) {
            stuffBytes++;
        }
    }
    const slipData = new Uint8Array(2 + stuffBytes + data.length);
    let offset = 0;
    slipData[offset++] = 0xc0;
    for (let i = 0; i < data.length; i++) {
        let c = data[i];
        if (c === 0xdb) {
            slipData[offset++] = 0xdb;
            slipData[offset++] = 0xdd;
        }
        else if (c === 0xc0) {
            slipData[offset++] = 0xdb;
            slipData[offset++] = 0xdc;
        }
        else {
            slipData[offset++] = c;
        }
    }
    slipData[offset++] = 0xc0;
    return slipData;
}
/**
 *  Decode %%data%% as SLIP encoded data, if valid SLIP-encoded data. Otherwise
 *  return ``null``.
 */
export function slipDecode(data) {
    const markers = [];
    for (let i = 0; i < data.length; i++) {
        if (data[i] === 0xc0) {
            markers.push(i);
        }
        if (markers.length < 2) {
            continue;
        }
        const result = [];
        for (let i = markers[0] + 1; i < markers[1]; i++) {
            if (data[i] === 0xdb) {
                if (data[i + 1] === 0xdc) {
                    result.push(0xc0);
                    i++;
                    continue;
                }
                else if (data[i + 1] === 0xdd) {
                    result.push(0xdb);
                    i++;
                    continue;
                }
            }
            result.push(data[i]);
        }
        return {
            data: new Uint8Array(result),
            start: markers[0],
            consumed: markers[1] + 1
        };
    }
    return null;
}
//# sourceMappingURL=slip.js.map