import { Device } from "./device.js";
import { hexlify } from "./utils.js";

import type { DeviceInfo } from "./device.js";

export const Magic = [ 0x6921506f, 0x1b31506f, 0x4881606f, 0x4361606f ];

export class DeviceEsp32c3 extends Device {
    readonly chipName = "ESP32-C3"

    readonly SPI_USR_OFFS = 0x18;
    //readonly SPI_USR1_OFFS = 0x1c;
    readonly SPI_USR2_OFFS = 0x20;
    readonly SPI_W0_OFFS = 0x58;

    readonly UART_DATE_REG_ADDR = 0x6000007c;

    // @TODO: Rename to spiRegisterAddress
    registerAddress(offset: number): number {
        return SPI_REG_BASE + offset;
    }

    // @TODO: Rename? setSpiLengthRegisters
    async _setDataLengths(mosiLength: number, misoLength: number): Promise<void> {
        const SPI_MOSI_DLEN_OFFS = 0x24;
        const SPI_MISO_DLEN_OFFS = 0x28;

        if (mosiLength > 0) {
            await this.writeSpiRegister(SPI_MOSI_DLEN_OFFS, mosiLength - 1);
        }
        if (misoLength > 0) {
            await this.writeSpiRegister(SPI_MISO_DLEN_OFFS, misoLength - 1);
        }
    }

    // ESP8266 maybe?
    /*
    _setDataLength(mosiLength: number, misoLength: number): Promise<void> {
        const SPI_DATA_LEN_REG = SPI_USR1_REG;
        const SPI_MOSI_BITLEN_S = 17;
        const SPI_MISO_BITLEN_S = 8;
        const mosiMask = mosiBits === 0 ? 0 : mosiBits - 1;
        const misoMask = misoBits === 0 ? 0 : misoBits - 1;
        const val = (misoMask << SPI_MISO_BITLEN_S) | (mosiMask << SPI_MOSI_BITLEN_S);
        await this.writeRegister(SPI_DATA_LEN_REG, val);
    }
    */

    async getDeviceInfo(): Promise<DeviceInfo> {
        const EFUSE_BASE = 0x60008800;
        const block1Addr = EFUSE_BASE + 0x044;
        const numWord = 3;
        const addr = block1Addr + 4 * numWord;

        const value = await this.readRegister(addr);

        const packageVersion = (value >> 21) & 0x07;
        const revision = (value >> 18) & 0x07;
        const description = `${ (packageVersion === 0) ? "": "unknown " }ESP32-C3 (revision: ${ revision })`;

        return { packageVersion, revision, description };
    }

    async getMacAddress(): Promise<string> {
        const MAC_EFUSE_REG = 0x60008800 + 0x044;

        const mac0 = await this.readRegister(MAC_EFUSE_REG) & 0xffffffff;
        const mac1 = (await this.readRegister(MAC_EFUSE_REG + 4)) & 0xffff;

        return [
            hexlify(mac1 >> 8, 1),
            hexlify(mac1 & 0xff, 1),
            hexlify(mac0 >> 24, 1),
            hexlify((mac0 >> 16) & 0xff, 1),
            hexlify((mac0 >> 8) & 0xff, 1),
            hexlify(mac0 & 0xff, 1),
        ].join(":");
    }

    checkMagic(magic: number) {
        return Magic.indexOf(magic) >= 0;
    }
}


const SPI_REG_BASE = 0x60002000;
