/**
 *  A simple communication library for the ESP Devices over the
 *  Serial Protocol for the UART bootloader.
 *
 *  See:
 *    ESPTool: 
 *    ESPTool-js: 
 *    Protocol: https://docs.espressif.com/projects/esptool/en/latest/esp32/advanced-topics/serial-protocol.html
 */

import { ESP_READ_REG, ESP_WRITE_REG } from "./constants.js";
import { Logger } from "./logger.js";
import { slipDecode, slipEncode } from "./slip.js";
import { CMD_SYNC, getErrorMessage, syncPacket } from "./protocol.js";
import {
    assert, concat, fromLeBytes, hexlify, stall, toLeBytes
} from "./utils.js";

import type { SerialPort } from "./serial.js";


export const Sequences: Record<string, string> = {
    Reset: "R 100 D 50 N",
    ResetUsb: "N 100 D 100 R 100 R 100 N",
    ResetHard: "R 100 N",
}

export interface DeviceInfo {
    packageVersion: number;
    revision: number;
    description: string;
}

export interface DeviceOptions {
    maxReadBuffer?: number;
    maxWriteBuffer?: number;
    logger?: Logger;
};

export type DeviceResponse = { value: number, data: Uint8Array };

/**
 *  The **BaseDevice** class is a minimal implementation of
 *  of the Serial Protocol necessary to read/write and detect
 *  the device magic number.
 */
export class BaseDevice {
    readonly serial: SerialPort;

    #readBuffer: Array<Uint8Array>;
    #writeBuffer: Array<Uint8Array>

    #maxReadBuffer: number;
    #maxWriteBuffer: number;

    readonly logger: Logger;

    constructor(serial: SerialPort, options?: DeviceOptions) {
        if (options == null) { options = { }; }

        this.serial = serial;
        this.logger = options.logger || Logger.voidLogger();

        this.#readBuffer = [ ];
        this.#writeBuffer = [ ];
        this.#maxReadBuffer = getValue("invalid options.maxReadBuffer", options.maxReadBuffer, 1 << 20);
        this.#maxWriteBuffer = getValue("invalid options.maxWriteBuffer", options.maxReadBuffer, 1 << 20);
    }

    get maxReadBuffer(): number { return this.#maxReadBuffer; }
    get maxWriteBuffer(): number { return this.#maxWriteBuffer; }

    get available(): number { return sum(this.#readBuffer); }
    get backlog(): number { return sum(this.#writeBuffer); }

    async connect(): Promise<number> {
        await this.serial.connect();

        await this.reset();
        await stall(50);
        await this.read();  // Flush

        await this.sync();
        await stall(100);
        await this.read();  // Flush

        return await this.readRegister(0x40001000);
    }

    /**
     *  Reads any data on the stream into the read buffer and returns
     *  the combined data.
     *
     *  Use [[_unread]] to place any data back on the read buffer
     *  to be processed in the future.
     */
    async read(): Promise<Uint8Array> {
        const input = await this.serial.read();
        if (input.length) {
            this.logger.data(input, "<<<");
            this.#readBuffer.push(input);
        }

        const result = concat(this.#readBuffer);
        this.#readBuffer = [ ];
        return result;
    }

    /**
     *  Read a SLIP packet, optionally matching the %%op%%. Returns
     *  ``null`` if no complete matching packet is found.
     *
     *  Any stray packets or bytes at the front of the read buffer
     *  are discarded.
     */
    async readSlipPacket(op?: number): Promise<null | DeviceResponse> {
        const data = await this.read();
        const slip = slipDecode(data);

        // No packet found; maybe we need more bytes
        if (slip == null) {
            this._unread(data);
            return null;
        }

        // Place unconsumed bytes back onto the read buffer
        this._unread(data.slice(slip.consumed));

        const resp = slip.data[0];
        const opRet = slip.data[1];
        const value = fromLeBytes(slip.data.slice(4, 8));
        const result = slip.data.slice(8);

        // @TODO: Should take the length into account; but esptool doesn't seem to

        if (resp === 1) {
            if (op == null || opRet === op) {
                return { value, data: result };
            }

            if (result[0] !== 0) {
                assert(false, getErrorMessage(result[1]), {
                    code: result[1]
                });
            }
        }

        throw new Error(`Bad response: ${ slip }`);
    }

    _unread(data: Uint8Array): void {
        this.#readBuffer.unshift(data);
    }

    async write(data: Uint8Array): Promise<boolean> {
        this.logger.data(data, ">>>");
        return this.serial.write(data);
    }

    async writeSlipPacket(data: Uint8Array): Promise<boolean> {
        return this.write(slipEncode(data));
    }

    /**
     *  Reset the connected device, optionally providing a reset
     *  %%sequence%% for fine control over the DTR/RTS signals.
     */
    async reset(sequence?: string): Promise<void> {
        if (sequence == null) { sequence = Sequences.ResetUsb; }

        this.logger.info(`Reset: ${ sequence }`);

        for (const cmd of sequence.split(/ /g)) {
            switch (cmd) {
                case "N":
                    await this.serial.signal({ });
                    break;
                case "D":
                    await this.serial.signal({ dtr: true });
                    break;
                case "R":
                    await this.serial.signal({ rts: true });
                    break;
                case "DR": case "RD":
                    await this.serial.signal({ dtr: true, rts: true });
                    break;
                default:
                    await stall(parseInt(cmd));
            }
        }
    }

    async sync(): Promise<any> {
        this.logger.info(`|>> Sync:`);

        const result = await this.command(CMD_SYNC, syncPacket());
        this.logger.assertDedent(result, "sync: no response from command");
        this.logger.info(`<<| (Sync) value=0x${ hexlify(result.value, 4) }`);
        return result;
    }

    /**
     *  Send a command to the connected device and parse the response.
     */
    async command(op: number, data?: Array<number> | Uint8Array, checksum?: number): Promise<null | DeviceResponse> {
        if (Array.isArray(data)) { data = new Uint8Array(data); }

        this.logger.info(`|>> Command: op=0x${ hexlify(op, 1) } data=0x${ data ? hexlify(data): "" }`);

        const packet = new Uint8Array(8 + (data ? data.length: 0));
        packet[0] = 0x00;
        packet[1] = op;

        if (data) {
            packet.set(toLeBytes(data.length, 2), 2);
            packet.set(data, 8);
        }

        if (checksum) { packet.set(toLeBytes(checksum, 4), 4); }

        await this.writeSlipPacket(packet);

        // Try reading up to 1000ms
        for (let i = 0; i < 100; i++) {
            const result = await this.readSlipPacket(op);
            if (result) {
                this.logger.info(`<<| (Command) value=0x${ hexlify(result.value, 4) } data=0x${ hexlify(result.data) }`);
                return result;
            }
            await stall(10);
        }

        this.logger.info(`<<| (Command) null`);

        return null;
    }

    async readRegister(address: number): Promise<number> {
        this.logger.info(`|>> ReadRegister: address=0x${ hexlify(address, 4) }`);

        const result = await this.command(ESP_READ_REG, toLeBytes(address, 4));
        this.logger.assertDedent(result, "readRegister: no response from command");
        this.logger.info(`<<| (ReadRegister) value=0x${ hexlify(result.value, 4) }`);

        return result.value;
    }
}


export abstract class Device extends BaseDevice {

    async connect(): Promise<number> {
        const magic = await super.connect();

        assert(this.checkMagic(magic), `invalid magic number: ${ magic}`, {
            magic
        });

        return magic;
    }

    async spiFlashCommand(command: number, data: Uint8Array, responseBits: number): Promise<number> {
        assert(responseBits <= 32, "max SPI response length is 32 bits", {
            length: responseBits
        });

        assert(data.length <= 64, "max SPI request length is 64 bytes", {
            length: data.length
        });

        const oldSpiUsr = await this.readSpiRegister(this.SPI_USR_OFFS);
        const oldSpiUsr2 = await this.readSpiRegister(this.SPI_USR2_OFFS);

        await this._setDataLengths(data.length * 8, responseBits);

        {
            let flags = SPI_USR_COMMAND;
            if (responseBits > 0) { flags |= SPI_USR_MISO; }
            if (data.length) { flags |= SPI_USR_MOSI; }
            await this.writeSpiRegister(this.SPI_USR_OFFS, flags);
        }

        {
            const val = (7 << SPI_USR2_COMMAND_LEN_SHIFT) | command;
            await this.writeSpiRegister(this.SPI_USR2_OFFS, val);
        }

        {
            let reg = this.SPI_W0_OFFS;

            if (data.length === 0) {
                await this.writeSpiRegister(reg, 0);
            } else {
                // TODO: I think this logic is wrong; copied mostly
                // from esptool-js, but the padding looks backwards
                if (data.length % 4 != 0) {
                    const padding = new Uint8Array(data.length % 4);
                    data = concat([ data, padding ]);
                }
                // TODO: This also looks wrong; like it stops short?
                for (let i = 0; i < data.length - 4; i += 4) {
                    await this.writeSpiRegister(reg, fromLeBytes(data.slice(i, i + 4)));
                    reg += 4;
                }
            }
        }

        await this.writeSpiRegister(0x00, SPI_CMD_USR);

        for (let i = 0; i < 11; i++) {
            const val = (await this.readSpiRegister(0x00)) & SPI_CMD_USR;
            if (val == 0) { break; }
            assert(i < 10, "SPI command did not complete in time");
        }

        const status = await this.readSpiRegister(this.SPI_W0_OFFS);

        await this.writeSpiRegister(this.SPI_USR_OFFS, oldSpiUsr);
        await this.writeSpiRegister(this.SPI_USR2_OFFS, oldSpiUsr2);

        return status;
    }

    async readSpiRegister(offset: number): Promise<number> {
        return await this.readRegister(this.registerAddress(offset));
    }

    async writeSpiRegister(offset: number, value: number, mask?: number, delayUs?: number, delayAfterUs?: number): Promise<void> {
        const address = this.registerAddress(offset);

        if (mask == null) { mask = 0xffffffff; }
        if (delayUs == null) { delayUs = 0; }

        this.logger.info(`|>> WriteRegister: address=0x${ hexlify(address, 4) } value=0x${ hexlify(value) } mask=0x${ hexlify(mask) } delayUs=${ delayUs } delayAfterUs=${ delayAfterUs }`);

        const fields = [ address, value, mask, delayUs ];
        if (delayAfterUs) {
            fields.push(this.UART_DATE_REG_ADDR, 0, 0, delayAfterUs);
        }

        const packet = concat(fields.map((v) => toLeBytes(v, 4)));

        const result = await this.command(ESP_WRITE_REG, packet);

        if (result) {
            this.logger.info(`<<| (WriteRegister) value=0x${ hexlify(result.value, 4) }`);
        } else {
            this.logger.info(`<<| (WriteRegister) null`);
        }
    }

    /**
     *  Returns true if this **Device** supports the %%magic%%
     *  number. This is used in connect to validate the expected
     *  devices matches the connected device.
     */
    abstract checkMagic(magic: number): boolean;

    abstract getDeviceInfo(): Promise<DeviceInfo>;

    abstract registerAddress(register: number): number;

    abstract readonly chipName: string;

    abstract readonly SPI_USR_OFFS: number;
    abstract readonly SPI_USR2_OFFS: number;
    abstract readonly SPI_W0_OFFS: number;

    abstract _setDataLengths(mosiLength: number, misoLength: number): Promise<void>;

    abstract readonly UART_DATE_REG_ADDR: number;
}

// SPI_USR register flags
const SPI_USR_COMMAND = 1 << 31;
const SPI_USR_MISO = 1 << 28;
const SPI_USR_MOSI = 1 << 27;

const SPI_CMD_USR = 1 << 18;

const SPI_USR2_COMMAND_LEN_SHIFT = 28;


function getValue(message: string, value: undefined | number, fallback: number): number {
    if (value == null) { return fallback; }
    assert(Number.isInteger(value), `${ message }: ${ value }`, { value });
    if (value < -1) { return -1; }
    return value;
}

function sum(values: Array<Uint8Array>): number {
    return values.reduce((accum, value) => (accum + value.length), 0);
}

