import { assert, hexlify } from "./utils.js";

function pad(c: string, length: number): string {
    while (c.length < length) { c += c; }
    return c.substring(0, length);
}

/*
const HEX = "0123456789abcdef";
function hex(data: Uint8Array): string {
    let result = "";
    for (let i = 0; i < data.length; i++) {
        const d = data[i];
        result += HEX[d >> 4] + HEX[d & 0x0f];
    }
    return result;
}
*/
export type Printer = (line: string) => void;

export class Logger {
    readonly printer: Printer;

    #indent: number;

    constructor(printer?: Printer) {
        this.printer = printer || console.log;
        this.#indent = 0;
    }

    get indent(): number { return this.#indent; }

    data(data: Uint8Array, title?: string): void {
        if (title == null) { title = "[DATA]"; }
        if (title) { title += " "; }

        this._print(`${ title }${ data.length } bytes`);

        for (let i = 0; i < data.length; i += 16) {
            const chunk = data.slice(i, i + 16);
            let line = "  " + hexlify(chunk.slice(0, 8)) + " " + hexlify(chunk.slice(8));
            line += pad(" ", 37 - line.length) + "|  ";
            for (let j = 0; j < chunk.length; j++) {
                const c = chunk[j];
                if (c >= 32 && c <= 127) {
                    line += String.fromCharCode(c);
                } else {
                    line += ".";
                }
            }
            this._print(line);
        }
    }

    info(message: string): void {
        this._print(message);
    }

    object(obj: any, title?: string): void {
        if (title == null) { title = "[DATA]"; }
        if (title) { title += " "; }

        this._print(`${ title }${ JSON.stringify(obj) }`);
    }

    _print(message: string): void {
        if (message.startsWith("<<|")) {
            message = message.substring(3).trim();
            this.#indent--;
        }

        let indent = message.startsWith("|>>");
        if (indent) {
            message = message.substring(3).trim();
        }

        this.printer(pad(" ", this.#indent * 2) + message);

        if (indent) { this.#indent++; }
    }

    /**
     *  Asserts %%cond%%, and dedents on failure.
     */
    assertDedent(cond: any, message: string, info?: Record<string, any>): asserts cond {
        if (cond) { return; }
        this.#indent--;
        assert(false, message, info);
    }

    /**
     *  A Logger which does not emit any output.
     */
    static voidLogger(): Logger {
        return new Logger((l: string) => { });
    }

    /**
     *  A Logger which logs to the console.
     */
    static logger(): Logger {
        return new Logger();
    }
}
