import fs from "fs";

import ioctl from "ioctl";

import { SerialPort as _SerialPort } from "./serial.js"
import { concat, stall } from "./utils.js";

const TIOCMGET = 0x4004746a, TIOCMSET = 0x8004746d;
const TIOCM_RTS = 4, TIOCM_DTR = 2;

export class SerialPort extends _SerialPort {
    readonly filename: string;
    #fd: null | number;

    constructor(filename: string) {
        super();
        this.filename = filename;
        this.#fd = null;
    }

    async connect(): Promise<void> {
        if (this.#fd != null) { throw new Error("already connected"); }
        this.#fd = fs.openSync(this.filename, fs.constants.O_RDWR | fs.constants.O_NONBLOCK);
    }

    async #getFd(): Promise<number> {
        if (this.#fd == null) { throw new Error("serial not connected; call connect first"); }
        return this.#fd;
    }

    async read(): Promise<Uint8Array> {
        const fd = await this.#getFd();

        const chunks = [ ];

        let length = 0
        while (true) {
            try {
                const buffer = new Uint8Array(1024);
                const l = fs.readSync(fd, buffer);
                chunks.push(buffer.slice(0, l));
                length += l;
                await stall(50);
            } catch (e: any) {
                if (e.code !== "EAGAIN") {
                    console.log("ERROR READ", e);
                    throw e;
                }
                break;
            }
        }

        return concat(chunks);
    }

    async getSignal(): Promise<{ dtr: boolean, rts: boolean }> {
        const fd = await this.#getFd();

        const result = Buffer.from([ 0 ]);
        ioctl(fd, TIOCMGET, result);
        return {
            dtr: !!(result[0] & TIOCM_DTR),
            rts: !!(result[0] & TIOCM_RTS),
        };
    }

    async signal(signal: { dtr?: boolean, rts?: boolean }): Promise<void> {
        const fd = await this.#getFd();

        const arg = Buffer.from([ 0 ]);
        if (signal.dtr) { arg[0] |= TIOCM_DTR; }
        if (signal.rts) { arg[0] |= TIOCM_RTS; }
        ioctl(fd, TIOCMSET, arg)
    }

    async write(data: Uint8Array): Promise<boolean> {
        const fd = await this.#getFd();

        fs.writeSync(fd, data);
        return true;
    }

}
/*
const decoder = new TextDecoder();


const serial = new SerialPort("/dev/cu.usbmodem1101");
serial.addListener("data", (data: Uint8Array) => {
    console.log("DATA", decoder.decode(data));
});

import { chipName } from "./chips.js";

(async function() {
    await serial.reset();
    await stall(50);
    await serial.read();

    await serial.sync();
    await stall(100);
    await serial.read();

    const result = await serial.readRegister(0x40001000);
    console.log("CHIP", chipName(result));
    await serial.read();

})();
*/
